
#include "CWVehicle.h"
#include "CarWorld.h"
#include "H_Object.h"
#include "H_Graphics.h"
#include "H_Variable.h"
#include "H_XML.h"

//CLASS Wheel:
Wheel::Wheel(
	CWVehicle *AVehicle,
	const Point3D &RelPos,
	REAL SpringStiffness,
	REAL MaxTravel,
	REAL WheelRadius,
	REAL DamperCompression,
	REAL DamperRebound,
	REAL Grip,
	REAL AccelFactor,
	REAL BrakeFactor,
	REAL SteerFactor,
	bool LockUp
) :
	m_Vehicle(AVehicle),
	RelPos(RelPos),
	SpringStiffness(SpringStiffness),
	MaxTravel(MaxTravel),
	WheelRadius(WheelRadius),
	DamperCompression(DamperCompression),
	DamperRebound(DamperRebound),
	Grip(Grip),
	AccelFactor(AccelFactor),
	BrakeFactor(BrakeFactor),
	SteerFactor(SteerFactor),
	LockUp(LockUp)
{
	reset();
}

void Wheel::reset()
{
//others:
	AverageLoad = this->m_Vehicle->Weight/4;

	Acceleration = Point3D(0,0,0);
	Angle = 0;
	DeltaTravel = 0;
	Travel = 0;
	Force = Point3D(0,0,0);
	Gs = 0;
	LandDirection = Point3D(0,0,0);
	LandSpeed = Point3D(0,0,0);
	Lock = false;
	Thrust = 0;
}
void Wheel::Set(const Ref &ARef, HInput *ACommand = NULL)
{
	MyRef = ARef;

	if (ACommand == NULL)
	{
		Thrust = 0;
		Lock = false;
	}
	else
	{
		REAL Gas = LIMIT(0.,ACommand->Yaxis,1.);
		REAL Brake = LIMIT(0.,-ACommand->Yaxis,1.);

		MyRef.Rotate(MyRef.GetUp()*(ACommand->Xaxis*SteerFactor));

		if (MyRef.Speed*MyRef.GetDirection()>0)
			Thrust = (Gas*AccelFactor - Brake*BrakeFactor);
		else //the gas pedal switches with the Brake
			Thrust = (Gas*BrakeFactor - Brake*AccelFactor);
		Lock = ACommand->button[HBRAKE_BUTTON];
	}

	//Hit the ground surface
	MyContact = m_Vehicle->m_CarWorld->m_Landscape->GetContact(GetRay());

	if (MyContact.Found)
	{
		Point3D ToGround = MyContact.Position-MyRef.Position;
		REAL OldTravel = Travel;
		Travel = LIMIT(0,MyRef.GetUp()*ToGround,-MaxTravel);
		MyRef.Position += MyRef.GetUp()*(Travel);
		DeltaTravel = (Travel-OldTravel)/Ref::TimeIncrement;

		//calculate the LandSpeed and LandDirection:
		if (Travel>0)
		{
			LandSpeed =
				MyContact.Normal^(MyRef.Speed^MyContact.Normal);
			//calc LandDirection:
			Point3D W = MyRef.GetUp()^MyRef.GetDirection();
			LandDirection = W^MyContact.Normal;
			LandDirection.normalize();
			Angle += (LandSpeed*LandDirection*Ref::TimeIncrement)/(2*WheelRadius);
		}
	}
	else
	{
		Travel = 0;
		LandSpeed.reset();
		LandDirection.reset();
	}
}

void Wheel::CalcForce()
{
	if (MyContact.Found)
	{
		REAL Load;
		//only apply force if the susp is under load:
		if (DeltaTravel>0)
			Load = (SpringStiffness*Travel
				+DamperCompression*DeltaTravel);
		else //DeltaTravel<0
			Load = (SpringStiffness*Travel
				+DamperRebound*DeltaTravel);

		if (Load>0)
		{
			//tanO = sinO/cosO
			REAL tanO = ((MyContact.Normal^(MyRef.GetUp())).norm())
				/(MyContact.Normal*MyRef.GetUp());
			tanO = LIMIT(0.f,tanO,1.f);
			Load = (1+tanO)*Load;
			Force = MyContact.Normal * Load;
		}
		else
			Force.reset();

		Load = H_MAX(0,Load);

		//calculate where the wheel is pushing:
		//laterally
		Point3D Traction = (LandDirection*(LandSpeed*LandDirection)-LandSpeed)*(Load);
		//Longitudinally:
		Traction += LandDirection*Thrust;
		if (LockUp && Lock)
			Traction = -LandSpeed*Load;
		//try too mutch traction and you slip:
		//max traction depends on load and grip
		//but the load is limited to 2*(Average load)
		//to prevent more than twice the normal traction
		//ex: at jump receptions
		if (Traction.norm()>min(Load,2*AverageLoad*Grip))
			Traction.normalize(Load*Grip);
		Force += Traction;
	}
	else
		Force.reset();
}
FixedVector &Wheel::GetRay()
{
	Ray.Position = MyRef.Position + MyRef.GetUp()*(-2*MaxTravel);
	Ray.Value = MyRef.GetUp();
	Ray.Value = -Ray.Value;
	return Ray;
}

void Wheel::draw()
{
	//Move to the wheel location an draw it
	Ref RenderRef = MyRef;
	RenderRef.Position += RenderRef.Z*WheelRadius;
	RenderRef.Rotate(RenderRef.X*(-Angle));
	MyOFFObject.draw(RenderRef);
}
void Wheel::writeXMLTag(ostream &out)
{
	out << "\t<CWWheel \n";
	//out << "\t\tMyOFFObject=\"" << MyOFFObject << "\"\n";
	out << "\t\tRelPos=\"" << RelPos << "\"\n";
	out << "\t\tSpringStiffness=\"" << SpringStiffness << "\"\n";
	out << "\t\tMaxTravel=\"" << MaxTravel << "\"\n";
	out << "\t\tWheelRadius=\"" << WheelRadius << "\"\n";
	out << "\t\tDamperCompression=\"" << DamperCompression << "\"\n";
	out << "\t\tDamperRebound=\"" << DamperRebound << "\"\n";
	out << "\t\tGrip=\"" << Grip << "\"\n";
	out << "\t\tAccelFactor=\"" << AccelFactor << "\"\n";
	out << "\t\tBrakeFactor=\"" << BrakeFactor << "\"\n";
	out << "\t\tSteerFactor=\"" << SteerFactor << "\"\n";
	out << "\t\tLockUp=\"" <<LockUp << "\">\n";
	out << "\t</CWWheel>\n";
}

//CLASS CWVehicle:
CWVehicle::CWVehicle(const char *name, HInput *Input) :
	m_Input(Input)//,
	/*OffFile("data/Vehicles/truck.off"),
	InitPos(0,3,2),
	ChassisPos(0,.1f,.5f),
	Mass(1000),
	Weight(Mass*EARTH_GRAVITY),
	MassDistrib(1500),
	RedLine(8000),
	MaxTorque(100),
	Friction(.5f)*/
{
	MyDriveTrain = DriveTrain(RedLine, MaxTorque);
	MyRef = InertRef(Mass, MassDistrib);

	if (m_Input == NULL) m_Input = new HInput();
	load(name);
}

void CWVehicle::load(const char *name)
{
	ifstream in(name, ios::in | ios::binary);

	map<string,HVariable*> DescribeVhc;
	DescribeVhc["OffFile"] =		new HVar<string>("OffFile",&OffFile);
	DescribeVhc["InitPos"] =		new HVar<Point3D>("InitPos",&InitPos);
	DescribeVhc["ChassisPos"] =		new HVar<Point3D>("ChassisPos",&ChassisPos);
	DescribeVhc["Mass"] =			new HVar<REAL>("Mass",&Mass);
	DescribeVhc["MassDistrib"] =	new HVar<REAL>("MassDistrib",&MassDistrib);
	DescribeVhc["RedLine"] =		new HVar<REAL>("RedLine",&RedLine);
	DescribeVhc["MaxTorque"] =		new HVar<REAL>("MaxTorque",&MaxTorque);
	DescribeVhc["Friction"] =		new HVar<REAL>("Friction",&Friction);

	Wheel W(this);
	map<string,HVariable*> DescribeWheel;
	DescribeWheel["RelPos"] =				new HVar<Point3D>("RelPos",&(W.RelPos));
	DescribeWheel["SpringStiffness"] =		new HVar<REAL>("SpringStiffness",&(W.SpringStiffness));
	DescribeWheel["MaxTravel"] =			new HVar<REAL>("MaxTravel",&(W.MaxTravel));
	DescribeWheel["WheelRadius"] =			new HVar<REAL>("WheelRadius",&(W.WheelRadius));
	DescribeWheel["DamperCompression"] =	new HVar<REAL>("DamperCompression",&(W.DamperCompression));
	DescribeWheel["DamperRebound"] =		new HVar<REAL>("DamperRebound",&(W.DamperRebound));
	DescribeWheel["Grip"] =					new HVar<REAL>("Grip",&(W.Grip));
	DescribeWheel["AccelFactor"] =			new HVar<REAL>("AccelFactor",&(W.AccelFactor));
	DescribeWheel["BrakeFactor"] =			new HVar<REAL>("BrakeFactor",&(W.BrakeFactor));
	DescribeWheel["SteerFactor"] =			new HVar<REAL>("SteerFactor",&(W.SteerFactor));
	DescribeWheel["LockUp"] =				new HVar<bool>("LockUp",&(W.LockUp));

	XmlTag tag;
	do
	{
		tag.read_from(in);
		switch (tag.get_type())
		{
		case XmlTag::broken:
			throw HException("parse error in vehicle file");
		case XmlTag::opening:
			{
				if (tag.name() == "CWVehicle")
				{
					tag.write_to(DescribeVhc);
					Weight = Mass*EARTH_GRAVITY;
				}
				if (tag.name() == "CWWheel")
				{
					W = Wheel(this);
					tag.write_to(DescribeWheel);
					Wheels.push_back(W);
				}
			}
		default:;
		}
	} while ((tag.get_type() != XmlTag::broken) && (tag.get_type() != XmlTag::empty));

	MyDriveTrain = DriveTrain(RedLine, MaxTorque);
	MyRef = InertRef(Mass, MassDistrib);
}

void CWVehicle::save(const char *name)
{
	ofstream out(name, ios::out);
	
	out << "<CWVehicle \n";
	out << "\tOffFile=\"" << OffFile << "\"\n";
	out << "\tInitPos=\"" << InitPos << "\"\n";
	out << "\tChassisPos=\"" << ChassisPos << "\"\n";
	out << "\tMass=\"" << Mass << "\"\n";
	out << "\tMassDistrib=\"" << MassDistrib << "\"\n";
	out << "\tRedLine=\"" << RedLine << "\"\n";
	out << "\tMaxTorque=\"" << MaxTorque << "\"\n";
	out << "\tFriction=\"" << Friction << "\">\n";

	for (vector<Wheel>::iterator I = Wheels.begin() ; I != Wheels.end() ; I++)
		(*I).writeXMLTag(out);

	out << "</CWVehicle>\n";
}

CWVehicle::~CWVehicle()
{
	delete m_Input;
}

void CWVehicle::reset()
{
	//cout << "in car.reset()\n";
	MyRef.reset();

	for (vector<Wheel>::iterator I = Wheels.begin() ; I != Wheels.end() ; I++)
		(*I).reset();

	//cout << "updating car params...\n";
	UpdateCWVehicleParams();
	//cout << "finnished resetting car...\n";
}

void CWVehicle::update()
{
	static Point3D Gravity(0,0,-EARTH_GRAVITY);

//Add the suspention forces:
	CalcForces();
	for (vector<Wheel>::iterator I = Wheels.begin() ; I != Wheels.end() ; I++)
		MyRef.Apply(FixedVector((*I).MyRef.Position, (*I).Force));
	MyRef.Apply(FixedVector(MyRef.Position, Gravity*MyRef.GetMass()));
	MyRef.Apply(FixedVector(MyRef.Position, MyRef.Speed*(-Friction*MyRef.Speed.norm())));

	MyRef.TimeClick();

	//make the rest of the car follow:
	m_Input->UpdateState();
	UpdateCWVehicleParams(*m_Input);
}

REAL CWVehicle::GetSpeed() const //returns speed in kph
{
	return MyRef.Speed.norm()*3.6f; // 3.6 = 3600/1000 to get kph
}

void CWVehicle::UpdateCWVehicleParams(HInput &ACommand)
{
	for (vector<Wheel>::iterator I = Wheels.begin() ; I != Wheels.end() ; I++)
		(*I).Set(
			MyRef.GetRef((*I).RelPos),
			&ACommand
		);
}
void CWVehicle::UpdateCWVehicleParams()
{
	for (vector<Wheel>::iterator I = Wheels.begin() ; I != Wheels.end() ; I++)
		(*I).Set(
			MyRef.GetRef((*I).RelPos)
		);
}
void CWVehicle::CalcForces()
{
	for (vector<Wheel>::iterator I = Wheels.begin() ; I != Wheels.end() ; I++)
		(*I).CalcForce();
}

void CWVehicle::draw_init()
{
	//MyOFFObject.readfile("data/Vehicles/vw.off", 2.5f);
	MyOFFObject.readfile("data/Vehicles/truck.off", 2.5f);
	MyOFFObject.SetShading(false);
	for (vector<Wheel>::iterator I = Wheels.begin() ; I != Wheels.end() ; I++)
		(*I).MyOFFObject.readfile("data/Vehicles/Wheel.off", .35f);
}

void CWVehicle::draw()
{
	if (Hgl::GetShadows()>0)
		ProjectShadow(m_CarWorld->LightDirection);
	drawShape();
}

void CWVehicle::drawShape()
{
	//draw the Wheels:
	for (vector<Wheel>::iterator I = Wheels.begin() ; I != Wheels.end() ; I++)
		(*I).draw();

	//draw the car:
	if (dynamic_cast<InCarCam*>(m_CarWorld->m_Camera)==NULL)
		MyOFFObject.draw(MyRef.GetRef(ChassisPos));
}

void CWVehicle::ProjectShadow(const Point3D &LightDirection)
{
	static Point3D ShadowPlane[3] = {Point3D(0,0,0),Point3D(1,0,0),Point3D(0,1,0)};

	Contact ShadowContact =
		m_CarWorld->m_Landscape->GetContact(FixedVector(MyRef.Position,-LightDirection));

	if (ShadowContact.Found)
	{
		Hgl::Enable(Hgl::SHADOW);
		glPushMatrix();
		Hgl::ShadowTransform(LightDirection,ShadowContact);
		drawShape();
		glPopMatrix();
		Hgl::Disable(Hgl::SHADOW);
	}
}

void CWVehicle::drawInfo()
{
	MyDriveTrain.drawInfo(MyRef.Speed.norm());
}

void CWVehicle::set_input(HInput *AInput)
{
	delete m_Input;
	m_Input = AInput;
}
