
#include "H_Standard.h"
#include "CarWorldClient.h"
#include "CarWorldServer.h"

#include <iostream>
#include "H_Variable.h"

static const char * const cfg_file_name = "data/config.cfg";

static const char * const help_text  =
	"available commands:\n"
	"help       : display this message\n"
	"version    : print version information\n"
	"set        : set a variable to a value\n"
	"exec       : execute configuration file\n"
	"bind       : bind a key to a command\n";

class ExecCFG : public HExecutable
{
public:
	ExecCFG(CarWorldClient *CWC) : CWC(CWC) {}
	void exec(const Command &c)
	{
		if (c.size()==2)
			CWC->execute_cfg(c[1].c_str());
		else
			cout << "usage: exec <file>\n";
	}
	virtual ~ExecCFG() {}
private:
	CarWorldClient *CWC;
};

class BindKey : public HExecutable
{
public:
	BindKey(CarWorldClient *CWC) : CWC(CWC) {}
	void exec(const Command &c)
	{
		if (c.size()==3)
			CWC->bind(KeyMap.find(c[1]),c[2].c_str());
		else
			cout << "usage: bind <key> \"<command line>\"\n";
	}
	void serialize(ostream &out)
	{
		map<H_KEY,string>::iterator I;
		for (I = CWC->KeyBindings.begin() ; I!=CWC->KeyBindings.end(); I++)
		{
			out << "bind " << KeyMap.find((*I).first) << " \"" << (*I).second << "\"\n";
		}
	}
	virtual ~BindKey() {}
private:
	CarWorldClient *CWC;
};

int find(int argc, char **argv, char *v)
{
	int i=0;
	while (i<argc && strcmp(argv[i],v)) i++;
	return i;
}

CarWorldClient::CarWorldClient(int argc, char **argv) :
	m_Hgl(NULL),
	IsPromptMode(false),
	m_Vehicle(NULL),
	m_CarWorld(NULL),
	width(600),
	height(400),
	full_screen(find(argc,argv,"-f")!=argc)
{
	cout << "starting " << name() << " " << version() << " ...\n";

	cout = hout;
	//cout << boolalpha;

	cout << "starting " << name() << " " << version() << " ...\n";
	m_CarWorld = new CarWorld(TimeRefreshRate());
	m_Vehicle = new CWVehicle("data/Vehicles/truck.vhc",HInput::GetBestInput(m_window,HInput::Key));
	m_CarWorld->add(m_Vehicle);

}

void CarWorldClient::draw_init()
{
	cout = hout;

	cout << "initiating graphics...\n";
	m_Hgl = new Hgl(m_window);
	m_Hgl->MakeCurrent();
	cout << endl;

	m_CarWorld->draw_init();
	cout << endl;

	Hgl::ThrowError(); //check to see if everything is OK
	
	cout << "\ninitiating command line parameters...\n";
	HExecutableSet *m_HExecutableSet = new HExecutableSet();
	m_HExecutableSet->add(new HVar<bool>("gl_finish", &Hgl::SetFinish,&Hgl::GetFinish));
	m_HExecutableSet->add(new HVar<bool>("gl_vertex_arrays", &Hgl::SetVertexArrays,&Hgl::GetVertexArrays));
	m_HExecutableSet->add(new HVar<bool>("gl_ext_compiled_vertex_array",&Hgl::SetExtCompiledVertexArrays,&Hgl::GetExtCompiledVertexArrays));
	m_HExecutableSet->add(new HVar<int>("gl_shadows", &Hgl::SetShadows, &Hgl::GetShadows));
	m_HExecutableSet->add(new HVar<Hgl::Enum>("gl_texturemode",&Hgl::SetTextureMode,&Hgl::GetTextureMode));
	m_HExecutableSet->add(new HVar<bool>("draw_background",&(m_CarWorld->draw_background)));
	m_HExecutableSet->add(new HVarObj<CarWorldClient,HInput::Type>("in_type", this, &CarWorldClient::set_in_type,&CarWorldClient::get_in_type));
	m_HExecutableSet->add(new HVarObj<CarWorldClient,int>("r_mode", this, &CarWorldClient::set_r_mode,&CarWorldClient::get_r_mode));

	m_Executables["set"] = m_HExecutableSet;
	m_Executables["bind"] = new BindKey(this);
	m_Executables["reset"] = new MethodCall<CarWorld>(m_CarWorld,&CarWorld::reset);
	m_Executables["next_camera"] = new MethodCall<CarWorld>(m_CarWorld,&CarWorld::next_camera);
	m_Executables["toggleconsole"] = new MethodCall<CarWorldClient>(this,&CarWorldClient::toggleconsole);
	m_Executables["help"] = new MethodCall<CarWorldClient>(this,&CarWorldClient::print_help);
	m_Executables["version"] = new MethodCall<CarWorldClient>(this,&CarWorldClient::print_version);
	m_Executables["exec"] = new ExecCFG(this);

	bind(HK_TAB,"toggleconsole");
	bind(HK_F2, "next_camera");
	bind(HK_F3, "reset");
	bind(HK_F4, "set in_type joystick");
	bind(HK_F5, "set in_type keyboard");

	execute_cfg(cfg_file_name);
}

void CarWorldClient::draw_shutdown()
{
	cout = hout;
	//state must be saved while graphics variables are still valid...
	ofstream cfg_file(cfg_file_name, ios::out);
	write_cfg(cfg_file);

	m_Hgl->MakeCurrent();
	m_CarWorld->draw_shutdown();
	delete m_Hgl;
}

CarWorldClient::~CarWorldClient()
{
	cout = hout;
	for (map<string,HExecutable *>::iterator I = m_Executables.begin(); I != m_Executables.end() ; I++)
		delete (*I).second;
	delete m_CarWorld;
	cout << name() << " " << version() << " terminated.\n";
}

unsigned int CarWorldClient::init_width() {return width;}

unsigned int CarWorldClient::init_height() {return height;}

bool CarWorldClient::init_full_screen() {return full_screen;}

const char *CarWorldClient::name()
{
	return "CarWorld";
}

const char *CarWorldClient::version() {return "v0.1.86 Marcus Hewat 1999";}

void CarWorldClient::toggleconsole()
{
	IsPromptMode = !IsPromptMode;
}

void CarWorldClient::execute_cfg(const char *FileName)
{
	cout << "executing " << FileName << "...\n";
	char buffer[1024];
	ifstream in(FileName, ios::in);
	if (!in)
	{
		cout << "could not open \"" << FileName << "\"\n";
	}
	else while (!in.eof())
	{
		in.getline(buffer,1024);
		CommandQ.push(buffer);
	}
}

void CarWorldClient::pars_command(const char *value)
{
	Command ACommand(value);
	try
	{
		//if nothing do nothing
		if (ACommand.size() <= 0) return;
		if (m_Executables.find(ACommand[0].c_str())==m_Executables.end())
			cout << "\"" << ACommand[0].c_str() << "\" : is not a valid executable\n";
		else
		{
			m_Executables[ACommand[0].c_str()]->exec(ACommand);
		}
	}
	catch (const exception &E)
	{
		cout << E.what() << endl;
	}
}

void CarWorldClient::bind(H_KEY key,const char *command)
{
	KeyBindings[key] = command;
}

void CarWorldClient::print_help()
{
	cout << help_text;
}

void CarWorldClient::print_version()
{
	cout << this->name() << " " << this->version() << endl;
	Hgl::PrintVersion(cout);
	Hgl::PrintExtentions(cout);
	PrintAllScreenModes(cout);
}

void CarWorldClient::write_cfg(ostream &out)
{
	for (map<string,HExecutable*>::iterator I = m_Executables.begin() ; I!=m_Executables.end() ; I++)
		(*I).second->serialize(out);
}

void CarWorldClient::set_in_type(HInput::Type t)
{
	m_Vehicle->set_input(HInput::GetBestInput(m_window,t));
}

HInput::Type CarWorldClient::get_in_type()
{
	return m_Vehicle->m_Input->get_type();
}

void CarWorldClient::set_r_mode(int mode)
{
	SetScreenMode(mode,cout);
}

int CarWorldClient::get_r_mode()
{
	//echo not supported yet...
	return 0;
}

void CarWorldClient::key_down(H_KEY AHKey, char c)
{
	cout = hout;
	if (IsPromptMode)
	{
		string ReturnedCommand(hout.HitKey(AHKey,c));
		if (!ReturnedCommand.empty())
			CommandQ.push(ReturnedCommand);
	}
	map<H_KEY,string>::iterator I = KeyBindings.find(AHKey);
	if (I != KeyBindings.end())
		pars_command((*I).second.c_str());
	//else
	//	cout << "\"" << KeyMap.find(AHKey) << "\" key unbound\n";
}

void CarWorldClient::resize(unsigned int width, unsigned int weight)
{
	cout = hout;
	m_Hgl->MakeCurrent();
	HRect ClientRect = HRect(0,0,(REAL)width,(REAL)weight);
	Hgl::ResizeWindow(ClientRect);
}

void CarWorldClient::on_idle(unsigned int elapsed_time)
{
	if (m_window != NULL)
	{
		cout = hout;
		m_CarWorld->update(elapsed_time);
		if (!CommandQ.empty())
		{
			pars_command(CommandQ.front().c_str());
			CommandQ.pop();
		}
		draw();
	}
}

void CarWorldClient::draw()
{
	if (m_window != NULL)
	{
		cout = hout;
		m_Hgl->MakeCurrent();
		m_CarWorld->draw();
		if (IsPromptMode)
			hout.draw();
		else
			m_CarWorld->DrawOnScreen();
		Hgl::Finish();
		Hgl::ThrowError();
		Hgl::SwapBuffers();
	}
}

int main(int argc, char **argv)
{
	cerr = herr;
	//Add(new CarWorldServer(argc, argv));
	//Add(new CarWorldClient(argc, argv));
	Add(new CarWorldClient(argc, argv));
	return EventLoop(argc, argv);
}
