
#include "H_Geometry.h"
#include "H_Input.h"

#define STEER_INCREMENT .1f

//CLASS HInput

//returns AHInputType input if possible or best alternative
HInput *HInput::GetBestInput(HWindow *w, Type type)
{
	HInput *ToReturn;
	switch (type)
	{
	case HInput::Joy :
#ifdef WIN32
		ToReturn = new DXJoystickInput(w);
		if(ToReturn->IsValid) break;
		delete ToReturn;
#endif //WIN32
	case HInput::Key :
		ToReturn = new KeyboardInput(w);
		if(ToReturn->IsValid) break;
		delete ToReturn;
	default :
		ToReturn = new HInput(w);
	}
	cout << "input device: " << ToReturn->GetDescription() << "\n";
	return ToReturn;
}

HInput::HInput(HWindow *w) : m_window(w)
{
	reset();
	IsValid = true;
}

void HInput::UpdateState() {}

HInput::Type HInput::get_type()
{
	return HInput::None;
}

const char *HInput::GetDescription()
{
	return "no description available for this input device.";
}
void HInput::reset()
{
	for (int i=0 ; i<JOYSTICK_BUTTON_NUMBER ; i++)
		button[i] = false;
	Xaxis = 0.;
	Yaxis = 0.;
}
HInput::~HInput()
{}

//CLASS HInput::Type
ostream &operator<<(ostream &out, const HInput::Type &t)
{
	switch (t)
	{
	case HInput::None :
		return out << "none";
	case HInput::Key :
		return out << "keyboard";
	case HInput::Joy :
		return out << "joystick";
	}
	return out;
}
istream &operator>>(istream &in, HInput::Type &t)
{
	string s;
	in >> s;
	if (s == "joystick")
		t = HInput::Joy;
	else if (s == "keyboard")
		t = HInput::Key;
	else
		t = HInput::None;
	return in;
}

//CLASS KeyboardInput:
KeyboardInput::KeyboardInput(HWindow *w) :
	HInput(w),
	UpKey(HK_UP),
	DownKey(HK_DOWN),
	LeftKey(HK_LEFT),
	RightKey(HK_RIGHT)
{
	ButtonKey[0] = HK_SPACE;
	reset();
	IsValid = true;
}

void KeyboardInput::UpdateState() 
{
    button[0] = m_window->IsPressed(ButtonKey[0]);

    bool UpState = m_window->IsPressed(UpKey);
    bool DownState = m_window->IsPressed(DownKey);
    bool LeftState = m_window->IsPressed(LeftKey);
    bool RightState = m_window->IsPressed(RightKey);

    if (LeftState) Xaxis = H_MIN(Xaxis+2*STEER_INCREMENT,1);
	else if (RightState)
		Xaxis = H_MAX(Xaxis-2*STEER_INCREMENT,-1);
	else if (Xaxis>0)
		Xaxis = H_MAX(0,Xaxis-STEER_INCREMENT);
	else if (Xaxis<0)
		Xaxis = H_MIN(0,Xaxis+STEER_INCREMENT);

    if (UpState)
		Yaxis = H_MIN(Yaxis+2*STEER_INCREMENT,1);
	else if (DownState)
		Yaxis = H_MAX(Yaxis-2*STEER_INCREMENT,-1);
	else if (Yaxis>0)
		Yaxis = H_MAX(0,Yaxis-STEER_INCREMENT);
	else if (Yaxis<0)
		Yaxis = H_MIN(0,Yaxis+STEER_INCREMENT);
}

HInput::Type KeyboardInput::get_type()
{
	return HInput::Key;
}

const char *KeyboardInput::GetDescription()
{
	return m_window->GetKeyboardDescription();
}

KeyboardInput::~KeyboardInput()
{}


//Windows 95/98 specific code useing directX (DirectInput)
#ifdef WIN32

#include "win32/H_Internals.h"

// constants used for scaling the input device
#define DEADZONE        300            // 3% of the axis range
#define RANGE_MAX       10000            // maximum positive axis value
#define RANGE_MIN       -10000           // minimum negative axis value




/**************** DXJoystickInput ******************/

DXJoystickInput::DXJoystickInput(HWindow *w) :
	HInput(w),
	gpdi(NULL),
	gpdiJoystick(NULL)
{
	HWinWindow *win = dynamic_cast<HWinWindow*>(w);
	IsValid = inputInitDirectInput(win->hInst, win->hWnd);

}

void DXJoystickInput::UpdateState()
{
	if (!IsValid) return;


    HRESULT     hRes;
    DIJOYSTATE  dijs;
    DWORD       dwInput = 0;

    // poll the joystick to read the current state
    hRes = gpdiJoystick->Poll();

    // read the device state
    hRes = gpdiJoystick->GetDeviceState(sizeof(DIJOYSTATE),
                                                &dijs);
    if(FAILED(hRes))
    {
        if((hRes == DIERR_INPUTLOST))
            inputAcquireDevices();
        // we did not read anything, return no motion
        return;
    }

	Xaxis = ((REAL)-dijs.lX)/RANGE_MAX;
	Yaxis = ((REAL)-dijs.lY)/RANGE_MAX;

	for (int i=0 ; i<JOYSTICK_BUTTON_NUMBER ; i++)
		button[i] = (dijs.rgbButtons[i] & 0x80) != 0;
}

HInput::Type DXJoystickInput::get_type()
{
	return HInput::Joy;
}

const char *DXJoystickInput::GetDescription()
{
	return JoysticProductName.c_str();
}

DXJoystickInput::~DXJoystickInput()
{
	if (!IsValid) return;


    OutputDebugString("Cleaning up after DirectInput\n");

    // Unacquire() and Release() device objects
    //
    // It is perfectly safe to call Unacquire() on a device that is not
    // currently acquired.  In fact, it is good practice to call
    // Unacquire() just prior to Release().
    if(gpdiJoystick)
    {
        gpdiJoystick->Unacquire();
        gpdiJoystick->Release();
        gpdiJoystick = NULL;
    }

    // Release() base object
    if(gpdi)
    {
        gpdi->Release();
        gpdi = NULL;
    }
}


/********************  direct input methods *********************/
/****this code was taken and modified from the DirectX 6.0 SDK***/
/********************by Marcus Hewat 1999************************/


//===========================================================================
// inputInitDirectInput
// Creates and initializes DirectInput objects
// Returns: true if succeeded
//===========================================================================

// constants used for scaling the input device
#define DEADZONE        300            // 3% of the axis range
#define RANGE_MAX       10000            // maximum positive axis value
#define RANGE_MIN       -10000           // minimum negative axis value


bool DXJoystickInput::inputInitDirectInput(HINSTANCE hInst, HWND hWnd)
{
    HRESULT             hRes;
    LPDIRECTINPUTDEVICE pdiTempDevice       = NULL;
    DIDEVCAPS           didc;

    // create the base DirectInput object
    hRes = DirectInputCreate(hInst, DIRECTINPUT_VERSION, &gpdi, NULL);
    if(FAILED(hRes))
    {
        cout << "Joystick support requires DirectX 5 or later.\n";
        return false;
    }

    // enumerate for joystick devices
    hRes = gpdi->EnumDevices(DIDEVTYPE_JOYSTICK,
                                    (LPDIENUMDEVICESCALLBACK)(&DXJoystickInput::inputEnumDeviceProc),
                                    this,
                                    DIEDFL_ATTACHEDONLY);
    if(FAILED(hRes))
    {
        cout << "EnumDevices() failed, joystick unavailable\n";
        return false;
    }

    // create a temporary "Device 1" object
    hRes = gpdi->CreateDevice(guidDevice, &pdiTempDevice, NULL);
    if(FAILED(hRes))
    {
        cout << "This option of ""CarWorld"" requires a joystick device.\n";
        return false;
    }

    // get a "Device 2" object
    // this is needed for access to the ForceFeedback functionality
    hRes = pdiTempDevice->QueryInterface(IID_IDirectInputDevice2,
                                                (void **)&gpdiJoystick);
    if(FAILED(hRes))
    {
        cout << "QueryInterface() failed, joystick unavailable\n";
        return false;
    }

    // we no longer need the temporary device, go ahead and release it.
    if(pdiTempDevice)
    {
        pdiTempDevice->Release();
        pdiTempDevice = NULL;
    }

    // set the device's data format
    //
    // This tells the device object to act like a specific device --
    // in our case, like a joystick
    hRes = gpdiJoystick->SetDataFormat(&c_dfDIJoystick);
    if(FAILED(hRes))
    {
		cout << "SetDataFormat() failed, joystick unavailable\n";
        return false;
    }

    // set the device's cooperative level
    //
    // ForceFeedback requires Exclusive access to the device.
    //hRes = gpdiJoystick->SetCooperativeLevel(hWnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND);
	hRes = gpdiJoystick->SetCooperativeLevel(NULL, DISCL_BACKGROUND | DISCL_NONEXCLUSIVE);
    if(FAILED(hRes))
    {
		cout << "SetCooperativeLevel() failed, joystick unavailable\n";
        return false;
    }

    // set joystick parameters (deadzone, etc)
    if(!inputPrepareDevice())
    {
        cout << "Device preparation failed, joystick unavailable\n";
        return false;
    }

    // get the device capabilities
    //
    // We're going to check to see if the device we created supports
    // ForceFeedback.  If so, we will create effects, if not, we'll
    // support standard joystick functionality
    didc.dwSize = sizeof(DIDEVCAPS);
    hRes = gpdiJoystick->GetCapabilities(&didc);
    if(FAILED(hRes))
    {
        cout << "GetCapabilities() failed, joystick unavailable\n";
        return false;
    }
    // if we get here, we succeeded
    return true;

} //*** end inputInitDirectInput()




// Enumerates DirectInput devices of type specified in call to
//  IDirectInput::EnumDevices()

bool CALLBACK DXJoystickInput::inputEnumDeviceProc(LPDIDEVICEINSTANCE pdidi, DXJoystickInput *This)
{
	//DXJoystickInput *This = (DXJoystickInput *)pv;
	This->JoysticProductName = pdidi->tszProductName;
    GUID *pguidDevice = NULL;

    // validate pv
    // BUGBUG

    // report back the instance guid of the device we enumerated
    This->guidDevice = pdidi->guidInstance;

    // BUGBUG for now, stop after the first device has been found
    return DIENUM_STOP;

} //*** end inputEnumDeviceProc()



// Acquires the input device(s).
bool DXJoystickInput::inputAcquireDevices()
{
    if(!gpdiJoystick)
    {
        return false;
    }

    // reacquire the device
    if(SUCCEEDED(gpdiJoystick->Acquire()))
    {
        // DirectInput automatically resets the device whenever
        // ownership changes, so we can assume we've got a device
        // unsullied by its previous owner.
        //inputCreateEffect(EF_BOUNCE | EF_FIRE | EF_EXPLODE);

        return true;
    }
    // if we get here, we did >not< acquire the device
    return false;

} //*** end inputAcquireDevices()



// Performs device preparation by setting the device's parameters (ie
// deadzone).
bool DXJoystickInput::inputPrepareDevice()
{
    HRESULT       hRes;
    DIPROPRANGE   dipr;
    DIPROPDWORD   dipdw;


    // quick check to make sure that the object pointer is non-NULL
    if(!gpdiJoystick)
        return false;

    // call Unacquire() on the device
    //
    // SetParameter() will fail if a device is currently acquired, we are
    // doing this here in case we get careless and forget to call this
    // function either before we call Acquire() or after we call Unacquire().
    gpdiJoystick->Unacquire();

    // set the axis ranges for the device
    //
    // We will use the same range for the X and Y axes.  We are setting them
    // fairly low since we are not concerned with anything other than
    // "left", "right", "forward", "backward" and "centered"
	//* prepare DIPROPRANGE structure
    dipr.diph.dwSize        = sizeof(DIPROPRANGE);
	dipr.diph.dwHeaderSize  = sizeof(dipr.diph);
	dipr.diph.dwHow         = DIPH_BYOFFSET;
	dipr.lMin               = RANGE_MIN;  // negative to the left/top
	dipr.lMax               = RANGE_MAX;  // positive to the right/bottom
    //* x-axis
    dipr.diph.dwObj         = DIJOFS_X;
    //* set the x-axis range property
    hRes = gpdiJoystick->SetProperty(DIPROP_RANGE, &dipr.diph);
    if(FAILED(hRes))
    {
        OutputDebugString("SetProperty(RANGE, X-Axis) failed.\n");
        return false;
    }
    //* y-axis
    dipr.diph.dwObj         = DIJOFS_Y;
    hRes = gpdiJoystick->SetProperty(DIPROP_RANGE, &dipr.diph);
    if(FAILED(hRes))
    {
        OutputDebugString("SetProperty(RANGE, Y-Axis) failed.\n");
        return false;
    }

    // set the deadzone for the device
    //
    // We will use the same deadzone percentage for the X and Y axes.
    // This call uses a symbolic constant for the deadzone percentage so that
    // it is easy to change if we decide we don't like it.
	//* prepare DIPROPDWORD structure
	dipdw.diph.dwSize       = sizeof(DIPROPDWORD);
	dipdw.diph.dwHeaderSize = sizeof(dipdw.diph);
	dipdw.diph.dwHow        = DIPH_BYOFFSET;
	dipdw.dwData            = DEADZONE;
    //* set the x-axis range property
    dipdw.diph.dwObj         = DIJOFS_X;
    hRes = gpdiJoystick->SetProperty(DIPROP_DEADZONE, &dipdw.diph);
    if(FAILED(hRes))
    {
        OutputDebugString("SetProperty(DEADZONE, X-Axis) failed.\n");
        return false;
    }
    //* y-axis
    dipdw.diph.dwObj         = DIJOFS_Y;
    hRes = gpdiJoystick->SetProperty(DIPROP_DEADZONE, &dipdw.diph);
    if(FAILED(hRes))
    {
        OutputDebugString("SetProperty(DEADZONE, Y-Axis) failed.\n");
        return false;
    }

    // Acquire the device(s)
    //
    // This is being done as a convenience since we unacquired earlier in
    // this function.  This does not guarantee that the device will be
    // acquired at the time we return from the function (in other words, we
    // are not going to spin here until we get a succeessful acquisition).
    inputAcquireDevices();

    // we've actually done somthing here
    return true;

} //** end inputPrepareDevice()

#endif //WIN32
